/*
 *  mutex.c
 *  Threads and Mutex
 *
 *  Bernhard Rabe
 *  17.11.2004
 */
 
#include <windows.h>
#include <stdio.h>

HANDLE hMutex=NULL;
HANDLE hEvent=NULL;

void printErrorAndExit(const char *msg, DWORD err,int id)
{
	LPSTR lpMsgBuf;
	if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
			          FORMAT_MESSAGE_FROM_SYSTEM | 
					  FORMAT_MESSAGE_IGNORE_INSERTS,
						NULL,
						err,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					  (LPTSTR) &lpMsgBuf,
					  0,
					  NULL ))
	{
		fprintf(stderr,"%s %d : %s\n",msg,id,lpMsgBuf);
		LocalFree(lpMsgBuf);
	}
	else
	{
		fprintf(stderr,"Error at FormatMesage: %d\n",err=GetLastError());
	}
	exit(err);
}


DWORD WINAPI Thread(LPVOID data)
{
    int id=*(int*)data,l;
    
    if(WAIT_OBJECT_0!=WaitForSingleObject(hEvent,INFINITE))
        printErrorAndExit("Error waiting for Event in",GetLastError(),id);
        
    if(WAIT_OBJECT_0!=WaitForSingleObject(hMutex,INFINITE))
        printErrorAndExit("Error accessing the Mutex in",GetLastError(),id);
        
    
    for(l=0;l<10;l++)
    {
        printf("%c",id);
        Sleep(1);
    }
    printf("\n");
    if(!ReleaseMutex(hMutex))
        printErrorAndExit("Error freeing the Mutex in",GetLastError(),id);
        
    return 0;    
}


int main(int argc, char *argv[])
{
    HANDLE hThread[10];
    int ids[10]={'-','*','+','#','~','$','','%','^',''};
    int i=2,j;
    
    if(argc>1)
    {
        i=atoi(argv[1]);
        if(i<2 || i> 10)
        {
            printf("Only 2-10 Threads allowed\n");
            printf("Usage: %s   - 2 Threads\nUsage: %s  [2-10] - 2-10 Threads\n");
            exit(1);
        }
    }
    
    hMutex=CreateMutex(NULL,TRUE,NULL);
    if(NULL==hMutex)
    {
        printErrorAndExit("Error creating the Mutex",GetLastError(),-1);
    }
    
    hEvent=CreateEvent(NULL,TRUE,FALSE,NULL);
    if(NULL==hEvent)
    {
        printErrorAndExit("Error creating the Event",GetLastError(),-1);
    }
    
    
    for(j=0;j<i;j++)
    {       //CREATE_SUSPENDED
        DWORD id;
        if(NULL==(hThread[j]=CreateThread(NULL,0,Thread,&ids[j],0, &id)))
            printErrorAndExit("Error creating Thread",GetLastError(),-1);
    }
    
    if(!ReleaseMutex(hMutex))
        printErrorAndExit("Error freeing the Mutex",GetLastError(),-1);
    
    Sleep(100);
    if(!SetEvent(hEvent))
        printErrorAndExit("Error signaling the Event",GetLastError(),-1);
    
    
    if(WaitForMultipleObjects(i,hThread,TRUE,INFINITE)==WAIT_FAILED)
        printErrorAndExit("Error waiting for Threads",GetLastError(),-1);
    
    return 0;
}